#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    List, Dict, Type
)

from ..exceptions import (
    CryptocurrencyError, SymbolError
)
from .adcoin import Adcoin
from .akashnetwork import AkashNetwork
from .algorand import Algorand
from .anon import Anon
from .aptos import Aptos
from .arbitum import Arbitrum
from .argoneum import Argoneum
from .artax import Artax
from .aryacoin import Aryacoin
from .asiacoin import Asiacoin
from .auroracoin import Auroracoin
from .avalanche import Avalanche
from .avian import Avian
from .axe import Axe
from .axelar import Axelar
from .bandprotocol import BandProtocol
from .base import Base
from .bata import Bata
from .beetlecoin import BeetleCoin
from .belacoin import BelaCoin
from .binance import Binance
from .bitcloud import BitCloud
from .bitcoin import Bitcoin
from .bitcoinatom import BitcoinAtom
from .bitcoincash import BitcoinCash
from .bitcoincashslp import BitcoinCashSLP
from .bitcoingold import BitcoinGold
from .bitcoingreen import BitcoinGreen
from .bitcoinplus import BitcoinPlus
from .bitcoinprivate import BitcoinPrivate
from .bitcoinsv import BitcoinSV
from .bitcoinz import BitcoinZ
from .bitcore import Bitcore
from .bitsend import BitSend
from .blackcoin import Blackcoin
from .blocknode import Blocknode
from .blockstamp import BlockStamp
from .bolivarcoin import Bolivarcoin
from .britcoin import BritCoin
from .canadaecoin import CanadaECoin
from .cannacoin import Cannacoin
from .cardano import Cardano
from .celo import Celo
from .chihuahua import Chihuahua
from .clams import Clams
from .clubcoin import ClubCoin
from .compcoin import Compcoin
from .cosmos import Cosmos
from .cpuchain import CPUChain
from .cranepay import CranePay
from .crave import Crave
from .dash import Dash
from .deeponion import DeepOnion
from .defcoin import Defcoin
from .denarius import Denarius
from .diamond import Diamond
from .digibyte import DigiByte
from .digitalcoin import Digitalcoin
from .divi import Divi
from .dogecoin import Dogecoin
from .dydx import dYdX
from .ecash import eCash
from .ecoin import ECoin
from .edrcoin import EDRCoin
from .egulden import eGulden
from .einsteinium import Einsteinium
from .elastos import Elastos
from .energi import Energi
from .eos import EOS
from .ergo import Ergo
from .ethereum import Ethereum
from .europecoin import EuropeCoin
from .evrmore import Evrmore
from .exclusivecoin import ExclusiveCoin
from .fantom import Fantom
from .feathercoin import Feathercoin
from .fetchai import FetchAI
from .filecoin import Filecoin
from .firo import Firo
from .firstcoin import Firstcoin
from .fix import FIX
from .flashcoin import Flashcoin
from .flux import Flux
from .foxdcoin import Foxdcoin
from .fujicoin import FujiCoin
from .gamecredits import GameCredits
from .gcrcoin import GCRCoin
from .gobyte import GoByte
from .gridcoin import Gridcoin
from .groestlcoin import GroestlCoin
from .gulden import Gulden
from .harmony import Harmony
from .helleniccoin import Helleniccoin
from .hempcoin import Hempcoin
from .horizen import Horizen
from .huobitoken import HuobiToken
from .hush import Hush
from .icon import Icon
from .injective import Injective
from .insanecoin import InsaneCoin
from .internetofpeople import InternetOfPeople
from .irisnet import IRISnet
from .ixcoin import IXCoin
from .jumbucks import Jumbucks
from .kava import Kava
from .kobocoin import Kobocoin
from .komodo import Komodo
from .landcoin import Landcoin
from .lbrycredits import LBRYCredits
from .linx import Linx
from .litecoin import Litecoin
from .litecoincash import LitecoinCash
from .litecoinz import LitecoinZ
from .lkrcoin import Lkrcoin
from .lynx import Lynx
from .mazacoin import Mazacoin
from .megacoin import Megacoin
from .metis import Metis
from .minexcoin import Minexcoin
from .monacoin import Monacoin
from .monero import Monero
from .monk import Monk
from .multiversx import MultiversX
from .myriadcoin import Myriadcoin
from .namecoin import Namecoin
from .nano import Nano
from .navcoin import Navcoin
from .near import Near
from .neblio import Neblio
from .neo import Neo
from .neoscoin import Neoscoin
from .neurocoin import Neurocoin
from .neutron import Neutron
from .newyorkcoin import NewYorkCoin
from .ninechronicles import NineChronicles
from .nix import NIX
from .novacoin import Novacoin
from .nubits import NuBits
from .nushares import NuShares
from .okcash import OKCash
from .oktchain import OKTChain
from .omni import Omni
from .onix import Onix
from .ontology import Ontology
from .optimism import Optimism
from .osmosis import Osmosis
from .particl import Particl
from .peercoin import Peercoin
from .pesobit import Pesobit
from .phore import Phore
from .pinetwork import PiNetwork
from .pinkcoin import Pinkcoin
from .pivx import Pivx
from .polygon import Polygon
from .poswcoin import PoSWCoin
from .potcoin import Potcoin
from .projectcoin import ProjectCoin
from .putincoin import Putincoin
from .qtum import Qtum
from .rapids import Rapids
from .ravencoin import Ravencoin
from .reddcoin import Reddcoin
from .ripple import Ripple
from .ritocoin import Ritocoin
from .rsk import RSK
from .rubycoin import Rubycoin
from .safecoin import Safecoin
from .saluscoin import Saluscoin
from .scribe import Scribe
from .secret import Secret
from .shadowcash import ShadowCash
from .shentu import Shentu
from .slimcoin import Slimcoin
from .smileycoin import Smileycoin
from .solana import Solana
from .solarcoin import Solarcoin
from .stafi import Stafi
from .stash import Stash
from .stellar import Stellar
from .stratis import Stratis
from .sugarchain import Sugarchain
from .sui import Sui
from .syscoin import Syscoin
from .terra import Terra
from .tezos import Tezos
from .theta import Theta
from .thoughtai import ThoughtAI
from .toacoin import TOACoin
from .tron import Tron
from .twins import TWINS
from .ultimatesecurecash import UltimateSecureCash
from .unobtanium import Unobtanium
from .vcash import Vcash
from .vechain import VeChain
from .verge import Verge
from .vertcoin import Vertcoin
from .viacoin import Viacoin
from .vivo import Vivo
from .voxels import Voxels
from .vpncoin import VPNCoin
from .wagerr import Wagerr
from .whitecoin import Whitecoin
from .wincoin import Wincoin
from .xinfin import XinFin
from .xuez import XUEZ
from .ycash import Ycash
from .zcash import Zcash
from .zclassic import ZClassic
from .zetacoin import Zetacoin
from .zilliqa import Zilliqa
from .zoobc import ZooBC
from .icryptocurrency import ICryptocurrency


class CRYPTOCURRENCIES:

    dictionary: Dict[str, Type[ICryptocurrency]] = {
        Adcoin.NAME: Adcoin,
        AkashNetwork.NAME: AkashNetwork,
        Algorand.NAME: Algorand,
        Anon.NAME: Anon,
        Aptos.NAME: Aptos,
        Arbitrum.NAME: Arbitrum,
        Argoneum.NAME: Argoneum,
        Artax.NAME: Artax,
        Aryacoin.NAME: Aryacoin,
        Asiacoin.NAME: Asiacoin,
        Auroracoin.NAME: Auroracoin,
        Avalanche.NAME: Avalanche,
        Avian.NAME: Avian,
        Axe.NAME: Axe,
        Axelar.NAME: Axelar,
        BandProtocol.NAME: BandProtocol,
        Base.NAME: Base,
        Bata.NAME: Bata,
        BeetleCoin.NAME: BeetleCoin,
        BelaCoin.NAME: BelaCoin,
        Binance.NAME: Binance,
        BitCloud.NAME: BitCloud,
        Bitcoin.NAME: Bitcoin,
        BitcoinAtom.NAME: BitcoinAtom,
        BitcoinCash.NAME: BitcoinCash,
        BitcoinCashSLP.NAME: BitcoinCashSLP,
        BitcoinGold.NAME: BitcoinGold,
        BitcoinGreen.NAME: BitcoinGreen,
        BitcoinPlus.NAME: BitcoinPlus,
        BitcoinPrivate.NAME: BitcoinPrivate,
        BitcoinSV.NAME: BitcoinSV,
        BitcoinZ.NAME: BitcoinZ,
        Bitcore.NAME: Bitcore,
        BitSend.NAME: BitSend,
        Blackcoin.NAME: Blackcoin,
        Blocknode.NAME: Blocknode,
        BlockStamp.NAME: BlockStamp,
        Bolivarcoin.NAME: Bolivarcoin,
        BritCoin.NAME: BritCoin,
        CanadaECoin.NAME: CanadaECoin,
        Cannacoin.NAME: Cannacoin,
        Cardano.NAME: Cardano,
        Celo.NAME: Celo,
        Chihuahua.NAME: Chihuahua,
        Clams.NAME: Clams,
        ClubCoin.NAME: ClubCoin,
        Compcoin.NAME: Compcoin,
        Cosmos.NAME: Cosmos,
        CPUChain.NAME: CPUChain,
        CranePay.NAME: CranePay,
        Crave.NAME: Crave,
        Dash.NAME: Dash,
        DeepOnion.NAME: DeepOnion,
        Defcoin.NAME: Defcoin,
        Denarius.NAME: Denarius,
        Diamond.NAME: Diamond,
        DigiByte.NAME: DigiByte,
        Digitalcoin.NAME: Digitalcoin,
        Divi.NAME: Divi,
        Dogecoin.NAME: Dogecoin,
        dYdX.NAME: dYdX,
        eCash.NAME: eCash,
        ECoin.NAME: ECoin,
        EDRCoin.NAME: EDRCoin,
        eGulden.NAME: eGulden,
        Einsteinium.NAME: Einsteinium,
        Elastos.NAME: Elastos,
        Energi.NAME: Energi,
        EOS.NAME: EOS,
        Ergo.NAME: Ergo,
        Ethereum.NAME: Ethereum,
        EuropeCoin.NAME: EuropeCoin,
        Evrmore.NAME: Evrmore,
        ExclusiveCoin.NAME: ExclusiveCoin,
        Fantom.NAME: Fantom,
        Feathercoin.NAME: Feathercoin,
        FetchAI.NAME: FetchAI,
        Filecoin.NAME: Filecoin,
        Firo.NAME: Firo,
        Firstcoin.NAME: Firstcoin,
        FIX.NAME: FIX,
        Flashcoin.NAME: Flashcoin,
        Flux.NAME: Flux,
        Foxdcoin.NAME: Foxdcoin,
        FujiCoin.NAME: FujiCoin,
        GameCredits.NAME: GameCredits,
        GCRCoin.NAME: GCRCoin,
        GoByte.NAME: GoByte,
        Gridcoin.NAME: Gridcoin,
        GroestlCoin.NAME: GroestlCoin,
        Gulden.NAME: Gulden,
        Harmony.NAME: Harmony,
        Helleniccoin.NAME: Helleniccoin,
        Hempcoin.NAME: Hempcoin,
        Horizen.NAME: Horizen,
        HuobiToken.NAME: HuobiToken,
        Hush.NAME: Hush,
        Icon.NAME: Icon,
        Injective.NAME: Injective,
        InsaneCoin.NAME: InsaneCoin,
        InternetOfPeople.NAME: InternetOfPeople,
        IRISnet.NAME: IRISnet,
        IXCoin.NAME: IXCoin,
        Jumbucks.NAME: Jumbucks,
        Kava.NAME: Kava,
        Kobocoin.NAME: Kobocoin,
        Komodo.NAME: Komodo,
        Landcoin.NAME: Landcoin,
        LBRYCredits.NAME: LBRYCredits,
        Linx.NAME: Linx,
        Litecoin.NAME: Litecoin,
        LitecoinCash.NAME: LitecoinCash,
        LitecoinZ.NAME: LitecoinZ,
        Lkrcoin.NAME: Lkrcoin,
        Lynx.NAME: Lynx,
        Mazacoin.NAME: Mazacoin,
        Megacoin.NAME: Megacoin,
        Metis.NAME: Metis,
        Minexcoin.NAME: Minexcoin,
        Monacoin.NAME: Monacoin,
        Monero.NAME: Monero,
        Monk.NAME: Monk,
        MultiversX.NAME: MultiversX,
        Myriadcoin.NAME: Myriadcoin,
        Namecoin.NAME: Namecoin,
        Nano.NAME: Nano,
        Navcoin.NAME: Navcoin,
        Near.NAME: Near,
        Neblio.NAME: Neblio,
        Neo.NAME: Neo,
        Neoscoin.NAME: Neoscoin,
        Neurocoin.NAME: Neurocoin,
        Neutron.NAME: Neutron,
        NewYorkCoin.NAME: NewYorkCoin,
        NineChronicles.NAME: NineChronicles,
        NIX.NAME: NIX,
        Novacoin.NAME: Novacoin,
        NuBits.NAME: NuBits,
        NuShares.NAME: NuShares,
        OKCash.NAME: OKCash,
        OKTChain.NAME: OKTChain,
        Omni.NAME: Omni,
        Onix.NAME: Onix,
        Ontology.NAME: Ontology,
        Optimism.NAME: Optimism,
        Osmosis.NAME: Osmosis,
        Particl.NAME: Particl,
        Peercoin.NAME: Peercoin,
        Pesobit.NAME: Pesobit,
        Phore.NAME: Phore,
        PiNetwork.NAME: PiNetwork,
        Pinkcoin.NAME: Pinkcoin,
        Pivx.NAME: Pivx,
        Polygon.NAME: Polygon,
        PoSWCoin.NAME: PoSWCoin,
        Potcoin.NAME: Potcoin,
        ProjectCoin.NAME: ProjectCoin,
        Putincoin.NAME: Putincoin,
        Qtum.NAME: Qtum,
        Rapids.NAME: Rapids,
        Ravencoin.NAME: Ravencoin,
        Reddcoin.NAME: Reddcoin,
        Ripple.NAME: Ripple,
        Ritocoin.NAME: Ritocoin,
        RSK.NAME: RSK,
        Rubycoin.NAME: Rubycoin,
        Safecoin.NAME: Safecoin,
        Saluscoin.NAME: Saluscoin,
        Scribe.NAME: Scribe,
        Secret.NAME: Secret,
        ShadowCash.NAME: ShadowCash,
        Shentu.NAME: Shentu,
        Slimcoin.NAME: Slimcoin,
        Smileycoin.NAME: Smileycoin,
        Solana.NAME: Solana,
        Solarcoin.NAME: Solarcoin,
        Stafi.NAME: Stafi,
        Stash.NAME: Stash,
        Stellar.NAME: Stellar,
        Stratis.NAME: Stratis,
        Sugarchain.NAME: Sugarchain,
        Sui.NAME: Sui,
        Syscoin.NAME: Syscoin,
        Terra.NAME: Terra,
        Tezos.NAME: Tezos,
        Theta.NAME: Theta,
        ThoughtAI.NAME: ThoughtAI,
        TOACoin.NAME: TOACoin,
        Tron.NAME: Tron,
        TWINS.NAME: TWINS,
        UltimateSecureCash.NAME: UltimateSecureCash,
        Unobtanium.NAME: Unobtanium,
        Vcash.NAME: Vcash,
        VeChain.NAME: VeChain,
        Verge.NAME: Verge,
        Vertcoin.NAME: Vertcoin,
        Viacoin.NAME: Viacoin,
        Vivo.NAME: Vivo,
        Voxels.NAME: Voxels,
        VPNCoin.NAME: VPNCoin,
        Wagerr.NAME: Wagerr,
        Whitecoin.NAME: Whitecoin,
        Wincoin.NAME: Wincoin,
        XinFin.NAME: XinFin,
        XUEZ.NAME: XUEZ,
        Ycash.NAME: Ycash,
        Zcash.NAME: Zcash,
        ZClassic.NAME: ZClassic,
        Zetacoin.NAME: Zetacoin,
        Zilliqa.NAME: Zilliqa,
        ZooBC.NAME: ZooBC
    }

    @classmethod
    def names(cls) -> List[str]:
        return list(cls.dictionary.keys())

    @classmethod
    def classes(cls) -> List[Type[ICryptocurrency]]:
        return list(cls.dictionary.values())

    @classmethod
    def cryptocurrency(cls, name: str) -> Type[ICryptocurrency]:

        if not cls.is_cryptocurrency(name=name):
            raise CryptocurrencyError(
                "Invalid cryptocurrency name", expected=cls.names(), got=name
            )

        return cls.dictionary[name]

    @classmethod
    def is_cryptocurrency(cls, name: str) -> bool:
        return name in cls.names()


def get_cryptocurrency(symbol: str) -> Type[ICryptocurrency]:
    for cls in CRYPTOCURRENCIES.classes():
        if symbol == cls.SYMBOL:
            return cls
    raise SymbolError(
        f"Cryptocurrency not found with this {symbol} symbol"
    )


__all__: List[str] = [
    "ICryptocurrency", "CRYPTOCURRENCIES", "get_cryptocurrency"
] + [
    cls.__name__ for cls in CRYPTOCURRENCIES.classes()
]



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, Union, Tuple
)
from Crypto.Hash import (
    SHA512, SHA3_256, keccak
)
from Crypto.Cipher import ChaCha20_Poly1305
from Crypto.Protocol.KDF import PBKDF2

import binascii
import crcmod.predefined
import hashlib
import hmac

from .libs.ripemd160 import ripemd160 as r160
from .consts import SLIP10_SECP256K1_CONST
from .utils import (
    get_bytes, encode, integer_to_bytes
)


def hmac_sha256(key: Union[bytes, str], data: Union[bytes, str]) -> bytes:
    """
    Generate an HMAC-SHA256 hash of the given data using the provided key.

    :param key: The key for the HMAC algorithm, as bytes or a string.
    :type key: Union[bytes, str]
    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]

    :return: The resulting HMAC-SHA256 hash as bytes.
    :rtype: bytes
    """

    if hasattr(hmac, "digest"):
        return hmac.digest(
            encode(key), encode(data), "sha256"
        )
    return hmac.new(
        encode(key), encode(data), hashlib.sha256
    ).digest()


def hmac_sha512(key: Union[bytes, str], data: Union[bytes, str]) -> bytes:
    """
    Generate an HMAC-SHA512 hash of the given data using the provided key.

    :param key: The key for the HMAC algorithm, as bytes or a string.
    :type key: Union[bytes, str]
    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]

    :return: The resulting HMAC-SHA512 hash as bytes.
    :rtype: bytes
    """

    if hasattr(hmac, "digest"):
        return hmac.digest(
            encode(key), encode(data), "sha512"
        )
    return hmac.new(
        encode(key), encode(data), hashlib.sha512
    ).digest()


def blake2b(data: Union[bytes, str], digest_size: int, key: Union[bytes, str] = b"", salt: Union[bytes, str] = b"") -> bytes:
    """
    Generate a BLAKE2b hash of the given data with optional key and salt.

    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]
    :param digest_size: The size of the resulting digest in bytes.
    :type digest_size: int
    :param key: Optional key for keyed hashing, as bytes or a string. Default is an empty byte string.
    :type key: Union[bytes, str]
    :param salt: Optional salt for added randomness, as bytes or a string. Default is an empty byte string.
    :type salt: Union[bytes, str], optional

    :return: The resulting BLAKE2b hash as bytes.
    :rtype: bytes
    """

    return hashlib.blake2b(
        encode(data), digest_size=digest_size, key=encode(key), salt=encode(salt)
    ).digest()


def blake2b_32(data: Union[bytes, str], key: Union[bytes, str] = b"", salt: Union[bytes, str] = b"") -> bytes:
    """
    Generate a 32-byte BLAKE2b hash of the given data with optional key and salt.

    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]
    :param key: Optional key for keyed hashing, as bytes or a string. Default is an empty byte string.
    :type key: Union[bytes, str]
    :param salt: Optional salt for added randomness, as bytes or a string. Default is an empty byte string.
    :type salt: Union[bytes, str], optional

    :return: The resulting 32-byte BLAKE2b hash as bytes.
    :rtype: bytes
    """

    return blake2b(data=data, digest_size=4, key=key, salt=salt)


def blake2b_40(data: Union[bytes, str], key: Union[bytes, str] = b"", salt: Union[bytes, str] = b"") -> bytes:
    """
    Generate a 40-byte BLAKE2b hash of the given data with optional key and salt.

    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]
    :param key: Optional key for keyed hashing, as bytes or a string. Default is an empty byte string.
    :type key: Union[bytes, str]
    :param salt: Optional salt for added randomness, as bytes or a string. Default is an empty byte string.
    :type salt: Union[bytes, str], optional

    :return: The resulting 40-byte BLAKE2b hash as bytes.
    :rtype: bytes
    """

    return blake2b(data=data, digest_size=5, key=key, salt=salt)


def blake2b_160(data: Union[bytes, str], key: Union[bytes, str] = b"", salt: Union[bytes, str] = b"") -> bytes:
    """
    Generate a 160-bit (20-byte) BLAKE2b hash of the given data with optional key and salt.

    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]
    :param key: Optional key for keyed hashing, as bytes or a string. Default is an empty byte string.
    :type key: Union[bytes, str]
    :param salt: Optional salt for added randomness, as bytes or a string. Default is an empty byte string.
    :type salt: Union[bytes, str], optional

    :return: The resulting 20-byte BLAKE2b hash as bytes.
    :rtype: bytes
    """

    return blake2b(data=data, digest_size=20, key=key, salt=salt)


def blake2b_224(data: Union[bytes, str], key: Union[bytes, str] = b"", salt: Union[bytes, str] = b"") -> bytes:
    """
    Generate a 224-bit (28-byte) BLAKE2b hash of the given data with optional key and salt.

    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]
    :param key: Optional key for keyed hashing, as bytes or a string. Default is an empty byte string.
    :type key: Union[bytes, str]
    :param salt: Optional salt for added randomness, as bytes or a string. Default is an empty byte string.
    :type salt: Union[bytes, str], optional

    :return: The resulting 28-byte BLAKE2b hash as bytes.
    :rtype: bytes
    """

    return blake2b(data=data, digest_size=28, key=key, salt=salt)


def blake2b_256(data: Union[bytes, str], key: Union[bytes, str] = b"", salt: Union[bytes, str] = b"") -> bytes:
    """
    Generate a 256-bit (32-byte) BLAKE2b hash of the given data with optional key and salt.

    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]
    :param key: Optional key for keyed hashing, as bytes or a string. Default is an empty byte string.
    :type key: Union[bytes, str]
    :param salt: Optional salt for added randomness, as bytes or a string. Default is an empty byte string.
    :type salt: Union[bytes, str], optional

    :return: The resulting 32-byte BLAKE2b hash as bytes.
    :rtype: bytes
    """

    return blake2b(data=data, digest_size=32, key=key, salt=salt)


def blake2b_512(data: Union[bytes, str], key: Union[bytes, str] = b"", salt: Union[bytes, str] = b"") -> bytes:
    """
    Generate a 512-bit (64-byte) BLAKE2b hash of the given data with optional key and salt.

    :param data: The data to be hashed, as bytes or a string.
    :type data: Union[bytes, str]
    :param key: Optional key for keyed hashing, as bytes or a string. Default is an empty byte string.
    :type key: Union[bytes, str]
    :param salt: Optional salt for added randomness, as bytes or a string. Default is an empty byte string.
    :type salt: Union[bytes, str], optional

    :return: The resulting 64-byte BLAKE2b hash as bytes.
    :rtype: bytes
    """

    return blake2b(data=data, digest_size=64, key=key, salt=salt)


def chacha20_poly1305_encrypt(
    key: Union[bytes, str], nonce: Union[bytes, str], assoc_data: Union[bytes, str], plain_text: Union[bytes, str]
) -> Tuple[bytes, bytes]:
    """
    Encrypt data using ChaCha20-Poly1305 AEAD cipher.

    :param key: The encryption key, as bytes or a string.
    :type key: Union[bytes, str]
    :param nonce: The nonce value, as bytes or a string.
    :type nonce: Union[bytes, str]
    :param assoc_data: The associated data to authenticate, as bytes or a string.
    :type assoc_data: Union[bytes, str]
    :param plain_text: The plaintext data to be encrypted, as bytes or a string.
    :type plain_text: Union[bytes, str]

    :return: A tuple containing the encrypted data and the authentication tag.
    :rtype: Tuple[bytes, bytes]
    """

    cipher: ChaCha20_Poly1305 = ChaCha20_Poly1305.new(
        key=encode(key), nonce=encode(nonce)
    )
    cipher.update(encode(assoc_data))
    return cipher.encrypt_and_digest(encode(plain_text))


def chacha20_poly1305_decrypt(
    key: Union[bytes, str], nonce: Union[bytes, str], assoc_data: Union[bytes, str], cipher_text: Union[bytes, str], tag: Union[bytes, str]
) -> bytes:
    """
    Decrypt data using ChaCha20-Poly1305 AEAD cipher.

    :param key: The decryption key, as bytes or a string.
    :type key: Union[bytes, str]
    :param nonce: The nonce value, as bytes or a string.
    :type nonce: Union[bytes, str]
    :param assoc_data: The associated data used during encryption, as bytes or a string.
    :type assoc_data: Union[bytes, str]
    :param cipher_text: The encrypted data to be decrypted, as bytes or a string.
    :type cipher_text: Union[bytes, str]
    :param tag: The authentication tag associated with the encrypted data, as bytes or a string.
    :type tag: Union[bytes, str]

    :return: The decrypted plaintext data.
    :rtype: bytes
    """

    cipher: ChaCha20_Poly1305 = ChaCha20_Poly1305.new(
        key=encode(key), nonce=encode(nonce)
    )
    cipher.update(encode(assoc_data))
    return cipher.decrypt_and_verify(encode(cipher_text), encode(tag))


def sha256(data: Union[str, bytes]) -> bytes:
    """
    Calculate the SHA-256 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The SHA-256 hash digest as bytes.
    :rtype: bytes
    """

    return hashlib.sha256(get_bytes(data)).digest()


def double_sha256(data: Union[str, bytes]) -> bytes:
    """
    Calculate the double SHA-256 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The double SHA-256 hash digest as bytes.
    :rtype: bytes
    """

    return hashlib.sha256(sha256(data)).digest()


def hash160(data: Union[str, bytes]) -> bytes:
    """
    Calculate the HASH160 hash (RIPEMD-160 of SHA-256) of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The HASH160 hash digest as bytes.
    :rtype: bytes
    """

    return ripemd160(sha256(data))


def crc32(data: Union[bytes, str]) -> bytes:
    """
    Calculate the CRC-32 checksum of the given data.

    :param data: The data to calculate the CRC-32 checksum for, as bytes or a string.
    :type data: Union[bytes, str]

    :return: The CRC-32 checksum as a 4-byte bytes object.
    :rtype: bytes
    """

    return integer_to_bytes(
        binascii.crc32(encode(data)), bytes_num=4
    )


def xmodem_crc(data: Union[bytes, str]) -> bytes:
    """
    Calculate the XMODEM CRC checksum of the given data.

    :param data: The data to calculate the XMODEM CRC checksum for, as bytes or a string.
    :type data: Union[bytes, str]

    :return: The XMODEM CRC checksum as a bytes object.
    :rtype: bytes
    """

    xmodem = crcmod.predefined.Crc("xmodem")
    return xmodem.new(encode(data)).digest()


def pbkdf2_hmac_sha512(
    password: Union[bytes, str], salt: Union[bytes, str], iteration_num: int, derived_key_length: Optional[int] = None
) -> bytes:
    """
    Derive a key using PBKDF2-HMAC-SHA512.

    :param password: The password to derive the key from, as bytes or a string.
    :type password: Union[bytes, str]
    :param salt: The salt value used in the key derivation process, as bytes or a string.
    :type salt: Union[bytes, str]
    :param iteration_num: The number of iterations of the HMAC-SHA512 hashing to apply.
    :type iteration_num: int
    :param derived_key_length: Optional. The desired length of the derived key in bytes.
                               If not provided, it defaults to SHA-512 digest size.
    :type derived_key_length: Optional[int]

    :return: The derived key as bytes.
    :rtype: bytes
    """

    if hasattr(hashlib, "pbkdf2_hmac"):
        return hashlib.pbkdf2_hmac("sha512", encode(password), encode(salt), iteration_num, derived_key_length)
    return PBKDF2(
        password, encode(salt), derived_key_length or SHA512.digest_size, count=iteration_num, hmac_hash_module=SHA512
    )


def kekkak256(data: Union[str, bytes]) -> bytes:
    """
    Calculate the Keccak-256 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The Keccak-256 hash digest as bytes.
    :rtype: bytes
    """

    return keccak.new(data=encode(data), digest_bits=256).digest()


def ripemd160(data: Union[str, bytes]) -> bytes:
    """
    Calculate the RIPEMD-160 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The RIPEMD-160 hash digest as bytes.
    :rtype: bytes
    """

    if "ripemd160" in hashlib.algorithms_available:
        hashlib.new("ripemd160", get_bytes(data)).digest()
    return r160(get_bytes(data))


def get_checksum(data: Union[str, bytes]) -> bytes:
    """
    Calculate the checksum for the given raw bytes.

    The checksum is derived by performing a double SHA-256 hash on the input
    and returning the first few bytes, as determined by `CHECKSUM_BYTE_LENGTH`.

    :param data: The raw data to checksum.
    :type data: Union[str, bytes]

    :returns: The checksum of the data.
    :rtype: bytes
    """

    return double_sha256(data)[:SLIP10_SECP256K1_CONST.CHECKSUM_BYTE_LENGTH]


def sha512(data: Union[str, bytes]) -> bytes:
    """
    Calculate the SHA-512 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The SHA-512 hash digest as bytes.
    :rtype: bytes
    """

    return hashlib.sha512(encode(data)).digest()


def sha512_256(data: Union[str, bytes]) -> bytes:
    """
    Calculate the SHA-512/256 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The SHA-512/256 hash digest as bytes.
    :rtype: bytes
    """

    if "sha512_256" in hashlib.algorithms_available:
        return hashlib.new("sha512_256", encode(data)).digest()
    return SHA512.new(encode(data), truncate="256").digest()


def sha3_256(data: Union[str, bytes]) -> bytes:
    """
    Calculate the SHA3-256 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The SHA3-256 hash digest as bytes.
    :rtype: bytes
    """

    if "sha3_256" in hashlib.algorithms_available:
        return hashlib.new("sha3_256", encode(data)).digest()
    return SHA3_256.new(encode(data)).digest()



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base58 import ensure_string
from ..libs.segwit_bech32 import (
    segwit_encode
)
from ..consts import PUBLIC_KEY_TYPES
from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Bitcoin
from ..crypto import sha256
from ..utils import (
    get_bytes, bytes_to_string
)
from .p2wpkh import P2WPKHAddress


class P2WSHAddress(P2WPKHAddress):

    witness_version: int = Bitcoin.NETWORKS.MAINNET.WITNESS_VERSIONS.P2WSH

    @staticmethod
    def name() -> str:
        """
        Return the name of the address type, which is "P2WSH".

        :return: The name of the address type.
        :rtype: str
        """

        return "P2WSH"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into a P2TR address.

        :param public_key: The public key to encode.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
            - public_key_type: Type of the public key (compressed or uncompressed).
            - hrp: Human-readable part (optional).
            - version: Address version (optional).
        :type kwargs: Any

        :return: The encoded P2TR address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        public_key_bytes: bytes = (
            public_key.raw_compressed()
            if kwargs.get("public_key_type", PUBLIC_KEY_TYPES.COMPRESSED) == PUBLIC_KEY_TYPES.COMPRESSED else
            public_key.raw_uncompressed()
        )
        script_hash: bytes = sha256(get_bytes(
            "5121" + bytes_to_string(public_key_bytes) + "51ae"
        ))
        return ensure_string(segwit_encode(
            kwargs.get("hrp", cls.hrp), kwargs.get("version", cls.witness_version), script_hash
        ))



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List

from .ed25519 import (
    SLIP10Ed25519ECC, SLIP10Ed25519Point, SLIP10Ed25519PublicKey, SLIP10Ed25519PrivateKey
)
from .ed25519.blake2b import (
    SLIP10Ed25519Blake2bECC, SLIP10Ed25519Blake2bPoint, SLIP10Ed25519Blake2bPublicKey, SLIP10Ed25519Blake2bPrivateKey
)
from .ed25519.monero import (
    SLIP10Ed25519MoneroECC, SLIP10Ed25519MoneroPoint, SLIP10Ed25519MoneroPublicKey, SLIP10Ed25519MoneroPrivateKey
)
from .nist256p1 import (
    SLIP10Nist256p1ECC, SLIP10Nist256p1Point, SLIP10Nist256p1PublicKey, SLIP10Nist256p1PrivateKey
)
from .secp256k1 import (
    SLIP10Secp256k1ECC, SLIP10Secp256k1Point, SLIP10Secp256k1PublicKey, SLIP10Secp256k1PrivateKey
)


__all__: List[str] = [
    "SLIP10Ed25519ECC", "SLIP10Ed25519Point", "SLIP10Ed25519PublicKey", "SLIP10Ed25519PrivateKey",
    "SLIP10Ed25519Blake2bECC", "SLIP10Ed25519Blake2bPoint", "SLIP10Ed25519Blake2bPublicKey", "SLIP10Ed25519Blake2bPrivateKey",
    "SLIP10Ed25519MoneroECC", "SLIP10Ed25519MoneroPoint", "SLIP10Ed25519MoneroPublicKey", "SLIP10Ed25519MoneroPrivateKey",
    "SLIP10Nist256p1ECC", "SLIP10Nist256p1Point", "SLIP10Nist256p1PublicKey", "SLIP10Nist256p1PrivateKey",
    "SLIP10Secp256k1ECC", "SLIP10Secp256k1Point", "SLIP10Secp256k1PublicKey", "SLIP10Secp256k1PrivateKey",
]




#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ecdsa.ecdsa import curve_secp256k1
from ecdsa.ellipticcurve import (
    Point, PointJacobi
)
from ecdsa import keys

import coincurve

from ....consts import SLIP10_SECP256K1_CONST
from ...iecc import IPoint
from ....utils import (
    bytes_to_integer, integer_to_bytes
)


class SLIP10Secp256k1PointCoincurve(IPoint):

    public_key: coincurve.PublicKey

    def __init__(self, public_key: coincurve.PublicKey) -> None:
        """
        Initialize the SLIP10Secp256k1PublicKey instance with a given public key.

        :param public_key: The public key object representing a Secp256k1 public key.
        :type public_key: coincurve.PublicKey

        :return: None
        """

        self.public_key = public_key

    @staticmethod
    def name() -> str:
        """
        Returns the name of the ecc class.

        :return: The name of the class.
        :rtype: str
        """

        return "SLIP10-Secp256k1"

    @classmethod
    def from_bytes(cls, point_bytes: bytes) -> IPoint:
        """
        Creates a point object from bytes.

        :param cls: The class object.
        :type cls: Type
        :param point_bytes: The bytes representing the point.
        :type point_bytes: bytes

        :return: A point object created from the provided bytes.
        :rtype: IPoint
        """

        if len(point_bytes) == SLIP10_SECP256K1_CONST.PUBLIC_KEY_UNCOMPRESSED_BYTE_LENGTH - 1:
            return cls(coincurve.PublicKey(SLIP10_SECP256K1_CONST.PUBLIC_KEY_UNCOMPRESSED_PREFIX + point_bytes))
        if len(point_bytes) == SLIP10_SECP256K1_CONST.PUBLIC_KEY_COMPRESSED_BYTE_LENGTH:
            return cls(coincurve.PublicKey(point_bytes))
        raise ValueError("Invalid point bytes")

    @classmethod
    def from_coordinates(cls, x: int, y: int) -> IPoint:
        """
        Creates a point object from x and y coordinates.

        :param cls: The class object.
        :type cls: Type
        :param x: The x-coordinate of the point.
        :type x: int
        :param y: The y-coordinate of the point.
        :type y: int

        :return: A point object representing the coordinates (x, y).
        :rtype: IPoint
        """

        try:
            return cls(coincurve.PublicKey.from_point(x, y))
        except ValueError as ex:
            raise ValueError("Invalid point coordinates") from ex

    def underlying_object(self) -> Any:
        """
        Returns the underlying public key object.

        :return: The underlying public key object.
        :rtype: Any
        """

        return self.public_key

    def x(self) -> int:
        """
        Returns the x of the point.

        :return: The x of the point.
        :rtype: int
        """

        return self.public_key.point()[0]

    def y(self) -> int:
        """
        Returns the y of the point.

        :return: The y of the point.
        :rtype: int
        """

        return self.public_key.point()[1]

    def raw_encoded(self) -> bytes:
        """
        Returns the encoded raw bytes representation of the public key, including any necessary prefix.

        :return: The encoded raw bytes of the public key.
        :rtype: bytes
        """

        return self.public_key.format(True)

    def raw_decoded(self) -> bytes:
        """
        Returns the decoded raw bytes representation of the public key, excluding any prefix.

        :return: The decoded raw bytes of the public key.
        :rtype: bytes
        """

        return self.public_key.format(False)[1:]

    def __add__(self, point: IPoint) -> IPoint:
        """
        Adds another point to the current public key point using elliptic curve point addition.

        :param point: The other point to add to the current point.
        :type point: IPoint

        :return: A new instance of the class with the resulting point after addition.
        :rtype: IPoint
        """

        return self.__class__(self.public_key.combine([point.underlying_object()]))

    def __radd__(self, point: IPoint) -> IPoint:
        """
        Adds the current public key point to another point using elliptic curve point addition.
        This method allows addition in reverse order, i.e., point + self.

        :param point: The other point to add to the current point.
        :type point: IPoint

        :return: A new instance of the class with the resulting point after addition.
        :rtype: IPoint
        """

        return self + point

    def __mul__(self, scalar: int) -> IPoint:
        """
        Multiplies the current public key point by a scalar using elliptic curve scalar multiplication.

        :param scalar: The scalar integer to multiply the point with.
        :type scalar: int

        :return: A new instance of the class with the resulting point after multiplication.
        :rtype: IPoint
        """

        bytes_num = None or ((scalar.bit_length() if scalar > 0 else 1) + 7) // 8
        return self.__class__(self.public_key.multiply(scalar.to_bytes(bytes_num, byteorder="big", signed=False)))

    def __rmul__(self, scalar: int) -> IPoint:
        """
        Performs scalar multiplication of the point.

        :param scalar: The scalar integer by which to multiply the point.
        :type scalar: int

        :return: Resulting point after scalar multiplication.
        :rtype: IPoint

        """

        return self * scalar


class SLIP10Secp256k1PointECDSA(IPoint):

    point: PointJacobi

    def __init__(self, point_obj: PointJacobi) -> None:
        """
        Initializes the class with a PointJacobi object.

        :param point_obj: The PointJacobi object representing the point.
        :type point_obj: PointJacobi
        """

        self.point = point_obj

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Secp256k1"

    @classmethod
    def from_bytes(cls, point_bytes: bytes) -> IPoint:
        """
        Creates a point from its byte representation.

        :param point_bytes: The byte representation of the point.
        :type point_bytes: bytes

        :return: An instance of IPoint representing the decoded point.
        :rtype: IPoint
        """

        try:
            return cls(
                PointJacobi.from_bytes(
                    curve_secp256k1, point_bytes
                )
            )
        except keys.MalformedPointError as ex:
            raise ValueError("Invalid point key bytes") from ex
        except AttributeError:
            return cls.from_coordinates(
                bytes_to_integer(point_bytes[:SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH]),
                bytes_to_integer(point_bytes[SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH:])
            )

    @classmethod
    def from_coordinates(cls, x: int, y: int) -> IPoint:
        """
        Creates a point from x and y.

        :param x: The x of the point.
        :type x: int
        :param y: The y of the point.
        :type y: int

        :return: An instance of IPoint representing the point with the given coordinates.
        :rtype: IPoint
        """

        return cls(
            PointJacobi.from_affine(
                Point(curve_secp256k1, x, y)
            )
        )

    def underlying_object(self) -> Any:
        """
        Returns the underlying point object.

        :return: The underlying point object.
        :rtype: Any
        """

        return self.point

    def x(self) -> int:
        """
        Returns the x of the point.

        :return: The x of the point.
        :rtype: int
        """

        return self.point.x()

    def y(self) -> int:
        """
        Returns the y of the point.

        :return: The y of the point.
        :rtype: int
        """

        return self.point.y()

    def raw_encoded(self) -> bytes:
        """
        Returns the raw encoded bytes of the point.

        :return: The raw encoded bytes of the point.
        :rtype: bytes
        """

        try:
            return self.point.to_bytes("compressed")
        except AttributeError:
            x: bytes = integer_to_bytes(self.point.x(), SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH)
            return b"\x03" + x if self.point.y() & 1 else b"\x02" + x

    def raw_decoded(self) -> bytes:
        """
        Returns the raw decoded bytes of the point.

        :return: The raw bytes of the point.
        :rtype: bytes
        """

        try:
            return self.point.to_bytes()
        except AttributeError:
            x: bytes = integer_to_bytes(self.point.x(), SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH)
            y: bytes = integer_to_bytes(self.point.y(), SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH)

            return x + y

    def __add__(self, point: IPoint) -> IPoint:
        """
        Performs addition with another point.

        :param point: The point to add.
        :type point: IPoint

        :return: A new instance of the class representing the result of the addition.
        :rtype: IPoint
        """

        return self.__class__(self.point + point.underlying_object())

    def __radd__(self, point: IPoint) -> IPoint:
        """
        Performs addition with another point in the context of scalar multiplication.

        :param point: The point to add (right operand).
        :type point: IPoint

        :return: A new instance of the class representing the result of the addition.
        :rtype: IPoint
        """

        return self + point

    def __mul__(self, scalar: int) -> IPoint:
        """
        Performs scalar multiplication on the point.

        :param scalar: The scalar value to multiply with the point.
        :type scalar: int

        :return: A new instance of the class representing the result of scalar multiplication.
        :rtype: IPoint
        """

        return self.__class__(self.point * scalar)

    def __rmul__(self, scalar: int) -> IPoint:
        """
        Performs scalar multiplication on the point.

        :param scalar: The scalar value to multiply with the point.
        :type scalar: int

        :return: The result of scalar multiplication.
        :rtype: IPoint
        """

        return self * scalar


if SLIP10_SECP256K1_CONST.USE == "coincurve":
    SLIP10Secp256k1Point = SLIP10Secp256k1PointCoincurve
elif SLIP10_SECP256K1_CONST.USE == "ecdsa":
    SLIP10Secp256k1Point = SLIP10Secp256k1PointECDSA
else:
    raise Exception(
        f"Invalid SLIP10-Secp256k1 use, (expected: 'coincurve' or 'ecdsa', got: '{SLIP10_SECP256K1_CONST.USE}')"
    )



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Union, Dict, List, Optional
)

import unicodedata

from ...entropies import (
    IEntropy, BIP39Entropy, BIP39_ENTROPY_STRENGTHS
)
from ...crypto import sha256
from ...exceptions import (
    Error, EntropyError, MnemonicError, ChecksumError
)
from ...utils import (
    get_bytes,
    bytes_to_binary_string,
    bytes_to_string,
    binary_string_to_integer,
    integer_to_binary_string,
    binary_string_to_bytes
)
from ..imnemonic import IMnemonic


class BIP39_MNEMONIC_WORDS:

    TWELVE: int = 12
    FIFTEEN: int = 15
    EIGHTEEN: int = 18
    TWENTY_ONE: int = 21
    TWENTY_FOUR: int = 24


class BIP39_MNEMONIC_LANGUAGES:

    CHINESE_SIMPLIFIED: str = "chinese-simplified"
    CHINESE_TRADITIONAL: str = "chinese-traditional"
    CZECH: str = "czech"
    ENGLISH: str = "english"
    FRENCH: str = "french"
    ITALIAN: str = "italian"
    JAPANESE: str = "japanese"
    KOREAN: str = "korean"
    PORTUGUESE: str = "portuguese"
    RUSSIAN: str = "russian"
    SPANISH: str = "spanish"
    TURKISH: str = "turkish"


class BIP39Mnemonic(IMnemonic):
    """
    Implements the BIP39 standard, allowing the creation of mnemonic phrases for
    generating deterministic keys, widely used across various cryptocurrencies.

    Here are available ``BIP39_MNEMONIC_WORDS``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | TWELVE                | 12                   |
    +-----------------------+----------------------+
    | FIFTEEN               | 15                   |
    +-----------------------+----------------------+
    | EIGHTEEN              | 18                   |
    +-----------------------+----------------------+
    | TWENTY_ONE            | 21                   |
    +-----------------------+----------------------+
    | TWENTY_FOUR           | 24                   |
    +-----------------------+----------------------+

    Here are available ``BIP39_MNEMONIC_LANGUAGES``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | CHINESE_SIMPLIFIED    | chinese-simplified   |
    +-----------------------+----------------------+
    | CHINESE_TRADITIONAL   | chinese-traditional  |
    +-----------------------+----------------------+
    | CZECH                 | czech                |
    +-----------------------+----------------------+
    | ENGLISH               | english              |
    +-----------------------+----------------------+
    | FRENCH                | french               |
    +-----------------------+----------------------+
    | ITALIAN               | italian              |
    +-----------------------+----------------------+
    | JAPANESE              | japanese             |
    +-----------------------+----------------------+
    | KOREAN                | korean               |
    +-----------------------+----------------------+
    | PORTUGUESE            | portuguese           |
    +-----------------------+----------------------+
    | RUSSIAN               | russian              |
    +-----------------------+----------------------+
    | SPANISH               | spanish              |
    +-----------------------+----------------------+
    | TURKISH               | turkish              |
    +-----------------------+----------------------+
    """

    word_bit_length: int = 11
    words_list_number: int = 2048
    words_list: List[int] = [
        BIP39_MNEMONIC_WORDS.TWELVE,
        BIP39_MNEMONIC_WORDS.FIFTEEN,
        BIP39_MNEMONIC_WORDS.EIGHTEEN,
        BIP39_MNEMONIC_WORDS.TWENTY_ONE,
        BIP39_MNEMONIC_WORDS.TWENTY_FOUR
    ]
    words_to_entropy_strength: Dict[int, int] = {
        BIP39_MNEMONIC_WORDS.TWELVE: BIP39_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT,
        BIP39_MNEMONIC_WORDS.FIFTEEN: BIP39_ENTROPY_STRENGTHS.ONE_HUNDRED_SIXTY,
        BIP39_MNEMONIC_WORDS.EIGHTEEN: BIP39_ENTROPY_STRENGTHS.ONE_HUNDRED_NINETY_TWO,
        BIP39_MNEMONIC_WORDS.TWENTY_ONE: BIP39_ENTROPY_STRENGTHS.TWO_HUNDRED_TWENTY_FOUR,
        BIP39_MNEMONIC_WORDS.TWENTY_FOUR: BIP39_ENTROPY_STRENGTHS.TWO_HUNDRED_FIFTY_SIX
    }
    languages: List[str] = [
        BIP39_MNEMONIC_LANGUAGES.CHINESE_SIMPLIFIED,
        BIP39_MNEMONIC_LANGUAGES.CHINESE_TRADITIONAL,
        BIP39_MNEMONIC_LANGUAGES.CZECH,
        BIP39_MNEMONIC_LANGUAGES.ENGLISH,
        BIP39_MNEMONIC_LANGUAGES.FRENCH,
        BIP39_MNEMONIC_LANGUAGES.ITALIAN,
        BIP39_MNEMONIC_LANGUAGES.JAPANESE,
        BIP39_MNEMONIC_LANGUAGES.KOREAN,
        BIP39_MNEMONIC_LANGUAGES.PORTUGUESE,
        BIP39_MNEMONIC_LANGUAGES.RUSSIAN,
        BIP39_MNEMONIC_LANGUAGES.SPANISH,
        BIP39_MNEMONIC_LANGUAGES.TURKISH
    ]
    wordlist_path: Dict[str, str] = {
        BIP39_MNEMONIC_LANGUAGES.CHINESE_SIMPLIFIED: "bip39/wordlist/chinese_simplified.txt",
        BIP39_MNEMONIC_LANGUAGES.CHINESE_TRADITIONAL: "bip39/wordlist/chinese_traditional.txt",
        BIP39_MNEMONIC_LANGUAGES.CZECH: "bip39/wordlist/czech.txt",
        BIP39_MNEMONIC_LANGUAGES.ENGLISH: "bip39/wordlist/english.txt",
        BIP39_MNEMONIC_LANGUAGES.FRENCH: "bip39/wordlist/french.txt",
        BIP39_MNEMONIC_LANGUAGES.ITALIAN: "bip39/wordlist/italian.txt",
        BIP39_MNEMONIC_LANGUAGES.JAPANESE: "bip39/wordlist/japanese.txt",
        BIP39_MNEMONIC_LANGUAGES.KOREAN: "bip39/wordlist/korean.txt",
        BIP39_MNEMONIC_LANGUAGES.PORTUGUESE: "bip39/wordlist/portuguese.txt",
        BIP39_MNEMONIC_LANGUAGES.RUSSIAN: "bip39/wordlist/russian.txt",
        BIP39_MNEMONIC_LANGUAGES.SPANISH: "bip39/wordlist/spanish.txt",
        BIP39_MNEMONIC_LANGUAGES.TURKISH: "bip39/wordlist/turkish.txt"
    }

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the mnemonic class.

        :return: The name of the entropy class.
        :rtype: str
        """
        return "BIP39"

    @classmethod
    def from_words(cls, words: int, language: str) -> str:
        """
        Generates a mnemonic phrase from a specified number of words.

        This method generates a mnemonic phrase based on the specified number of words and language.

        :param words: The number of words for the mnemonic phrase.
        :type words: int
        :param language: The language for the mnemonic phrase.
        :type language: str

        :return: The generated mnemonic phrase.
        :rtype: str
        """
        if words not in cls.words_list:
            raise MnemonicError("Invalid mnemonic words number", expected=cls.words_list, got=words)

        return cls.from_entropy(
            entropy=BIP39Entropy.generate(cls.words_to_entropy_strength[words]), language=language
        )

    @classmethod
    def from_entropy(cls, entropy: Union[str, bytes, IEntropy], language: str, **kwargs) -> str:
        """
        Generates from entropy data.

        :param entropy: The entropy data used to generate the mnemonic phrase.
        :type entropy: Union[str, bytes, IEntropy]
        :param language: The language for the mnemonic phrase.
        :type language: str

        :return: The generated mnemonic phrase.
        :rtype: str
        """
        if isinstance(entropy, str) or isinstance(entropy, bytes):
            return cls.encode(entropy=entropy, language=language)
        elif isinstance(entropy, BIP39Entropy):
            return cls.encode(entropy=entropy.entropy(), language=language)
        raise EntropyError(
            "Invalid entropy instance", expected=[str, bytes, BIP39Entropy], got=type(entropy)
        )

    @classmethod
    def encode(cls, entropy: Union[str, bytes], language: str) -> str:
        """
        Encodes entropy into a mnemonic phrase.

        This method converts a given entropy value into a mnemonic phrase according to the specified language.

        :param entropy: The entropy to encode into a mnemonic phrase.
        :type entropy: Union[str, bytes]
        :param language: The language for the mnemonic phrase.
        :type language: str

        :return: The encoded mnemonic phrase.
        :rtype: str
        """

        entropy: bytes = get_bytes(entropy)
        if not BIP39Entropy.is_valid_bytes_strength(len(entropy)):
            raise EntropyError(
                "Wrong entropy strength", expected=BIP39Entropy.strengths, got=(len(entropy) * 8)
            )

        entropy_binary_string: str = bytes_to_binary_string(get_bytes(entropy), len(entropy) * 8)
        entropy_hash_binary_string: str = bytes_to_binary_string(sha256(entropy), 32 * 8)
        mnemonic_bin: str = entropy_binary_string + entropy_hash_binary_string[:len(entropy) // 4]

        mnemonic: List[str] = []
        words_list: List[str] = cls.normalize(cls.get_words_list_by_language(language=language))
        if len(words_list) != cls.words_list_number:
            raise Error(
                "Invalid number of loaded words list", expected=cls.words_list_number, got=len(words_list)
            )

        for index in range(len(mnemonic_bin) // cls.word_bit_length):
            word_bin: str = mnemonic_bin[index * cls.word_bit_length:(index + 1) * cls.word_bit_length]
            word_index: int = binary_string_to_integer(word_bin)
            mnemonic.append(words_list[word_index])

        return " ".join(cls.normalize(mnemonic))

    @classmethod
    def decode(
        cls, mnemonic: str, checksum: bool = False, words_list: Optional[List[str]] = None, words_list_with_index: Optional[dict] = None
    ) -> str:
        """
        Decodes a mnemonic phrase into its corresponding entropy.

        This method converts a given mnemonic phrase back into its original entropy value.
        It also verifies the checksum to ensure the mnemonic is valid.

        :param mnemonic: The mnemonic phrase to decode.
        :type mnemonic: str
        :param checksum: Whether to include the checksum in the returned entropy.
        :type checksum: bool
        :param words_list: Optional list of words used to decode the mnemonic. If not provided, the method will use the default word list for the language detected.
        :type words_list: Optional[List[str]]
        :param words_list_with_index: Optional dictionary mapping words to their indices for decoding. If not provided, the method will use the default mapping.
        :type words_list_with_index: Optional[dict]

        :return: The decoded entropy as a string.
        :rtype: str
        """

        words: list = cls.normalize(mnemonic)
        if len(words) not in cls.words_list:
            raise MnemonicError("Invalid mnemonic words count", expected=cls.words_list, got=len(words))

        if not words_list or not words_list_with_index:
            words_list, language = cls.find_language(mnemonic=words)
            if len(words_list) != cls.words_list_number:
                raise Error(
                    "Invalid number of loaded words list", expected=cls.words_list_number, got=len(words_list)
                )
            words_list_with_index: dict = {
                words_list[i]: i for i in range(len(words_list))
            }

        if len(words_list) != cls.words_list_number:
            raise Error(
                "Invalid number of loaded words list", expected=cls.words_list_number, got=len(words_list)
            )

        mnemonic_bin: str = "".join(map(
            lambda word: integer_to_binary_string(
                words_list_with_index[word], cls.word_bit_length
            ), words
        ))

        mnemonic_bit_length: int = len(mnemonic_bin)
        checksum_length: int = mnemonic_bit_length // 33
        checksum_bin: str = mnemonic_bin[-checksum_length:]
        entropy: bytes = binary_string_to_bytes(
            mnemonic_bin[:-checksum_length], checksum_length * 8
        )
        entropy_hash_bin: str = bytes_to_binary_string(
            sha256(entropy), 32 * 8
        )
        checksum_bin_got: str = entropy_hash_bin[:checksum_length]
        if checksum_bin != checksum_bin_got:
            raise ChecksumError(
                "Invalid checksum", expected=checksum_bin, got=checksum_bin_got
            )

        if checksum:
            pad_bit_len: int = (
                mnemonic_bit_length
                if mnemonic_bit_length % 8 == 0 else
                mnemonic_bit_length + (8 - mnemonic_bit_length % 8)
            )
            return bytes_to_string(
                binary_string_to_bytes(mnemonic_bin, pad_bit_len // 4)
            )
        return bytes_to_string(entropy)

    @classmethod
    def is_valid(
        cls,
        mnemonic: Union[str, List[str]],
        words_list: Optional[List[str]] = None,
        words_list_with_index: Optional[dict] = None
    ) -> bool:
        """
        Validates a mnemonic phrase.

        This method checks whether the provided mnemonic phrase is valid by attempting to decode it.
        If the decoding is successful without raising any errors, the mnemonic is considered valid.

        :param mnemonic: The mnemonic phrase to validate. It can be a string or a list of words.
        :type mnemonic: Union[str, List[str]]
        :param words_list: Optional list of words to be used for validation. If not provided, the method will use the default word list.
        :type words_list: Optional[List[str]]
        :param words_list_with_index: Optional dictionary mapping words to their indices for validation. If not provided, the method will use the default mapping.
        :type words_list_with_index: Optional[dict]

        :return: True if the mnemonic phrase is valid, False otherwise.
        :rtype: bool
        """

        try:
            cls.decode(
                mnemonic=mnemonic, words_list=words_list, words_list_with_index=words_list_with_index
            )
            return True
        except (Error, KeyError):
            return False

    @classmethod
    def normalize(cls, mnemonic: Union[str, List[str]]) -> List[str]:
        """
        Normalizes the given mnemonic by splitting it into a list of words if it is a string.

        :param mnemonic: The mnemonic value, which can be a single string of words or a list of words.
        :type mnemonic: Union[str, List[str]]

        :return: A list of words from the mnemonic.
        :rtype: List[str]
        """

        mnemonic: list = mnemonic.split() if isinstance(mnemonic, str) else mnemonic
        return list(map(lambda _: unicodedata.normalize("NFKD", _.lower()), mnemonic))



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Type, Union, Optional
)

from ..cryptocurrencies import Bitcoin
from ..eccs import IEllipticCurveCryptography
from ..consts import PUBLIC_KEY_TYPES
from ..addresses import P2WPKHInP2SHAddress
from ..exceptions import DerivationError
from ..derivations import (
    IDerivation, BIP49Derivation
)
from .bip44 import BIP44HD


class BIP49HD(BIP44HD):

    _derivation: BIP49Derivation

    def __init__(
        self, ecc: Type[IEllipticCurveCryptography], public_key_type: str = PUBLIC_KEY_TYPES.COMPRESSED, **kwargs
    ) -> None:
        """
        Initializes a BIP49HD object.

        :param ecc: Elliptic curve cryptography implementation.
        :type ecc: Type[IEllipticCurveCryptography]
        :param public_key_type: Type of public key to use (default: COMPRESSED).
        :type public_key_type: str

        :param kwargs: Additional keyword arguments for initialization.
        """

        super(BIP49HD, self).__init__(ecc=ecc, public_key_type=public_key_type, **kwargs)

        self._derivation = BIP49Derivation(
            coin_type=kwargs.get("coin_type", 0),
            account=kwargs.get("account", 0),
            change=kwargs.get("change", "external-chain"),
            address=kwargs.get("address", 0)
        )

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the bip class.

        :return: The name of the bip class.
        :rtype: str
        """
        return "BIP49"

    def from_derivation(self, derivation: IDerivation) -> "BIP49HD":
        """
        Sets the derivation path for the BIP49HD instance from a given derivation object.

        :param derivation: The derivation object to set.
        :type derivation: IDerivation

        :return: The updated BIP49HD instance.
        :rtype: BIP49HD
        """

        if not isinstance(derivation, BIP49Derivation):
            raise DerivationError(
                "Invalid derivation instance", expected=BIP49Derivation, got=type(derivation)
            )

        self.clean_derivation()
        self._derivation = derivation
        for index in self._derivation.indexes():
            self.drive(index)
        return self

    def root_xprivate_key(
        self, version: Union[bytes, int] = Bitcoin.NETWORKS.MAINNET.XPRIVATE_KEY_VERSIONS.P2WPKH_IN_P2SH, encoded: bool = True
    ) -> Optional[str]:
        """
        Generates the root extended private key (xprivate key) for the BIP44HD instance.

        :param version: The version bytes or integer for the xprivate key.
        :type version: Union[bytes, int]
        :param encoded: Whether to encode the xprivate key or return it in raw form, defaults to True.
        :type encoded: bool

        :return: The root extended private key (xprivate key) if available, otherwise None.
        :rtype: Optional[str]
        """

        return super(BIP44HD, self).root_xprivate_key(version=version, encoded=encoded)

    def root_xpublic_key(
        self, version: Union[bytes, int] = Bitcoin.NETWORKS.MAINNET.XPUBLIC_KEY_VERSIONS.P2WPKH_IN_P2SH, encoded: bool = True
    ) -> Optional[str]:
        """
        Generates the root extended public key (xpublic key) for the BIP44HD instance.

        :param version: The version bytes or integer for the xpublic key.
        :type version: Union[bytes, int]
        :param encoded: Whether to encode the xpublic key or return it in raw form, defaults to True.
        :type encoded: bool

        :return: The root extended public key (xpublic key) if available, otherwise None.
        :rtype: Optional[str]
        """

        return super(BIP44HD, self).root_xpublic_key(version=version, encoded=encoded)

    def xprivate_key(
        self, version: Union[bytes, int] = Bitcoin.NETWORKS.MAINNET.XPRIVATE_KEY_VERSIONS.P2WPKH_IN_P2SH, encoded: bool = True
    ) -> Optional[str]:
        """
            Generates the extended private key (xprivate key) for the BIP44HD instance.

            :param version: The version bytes or integer for the xprivate key.
            :type version: Union[bytes, int]
            :param encoded: Whether to encode the xprivate key or return it in raw form, defaults to True.
            :type encoded: bool

            :return: The extended private key (xprivate key) if available, otherwise None.
            :rtype: Optional[str]
            """

        return super(BIP44HD, self).xprivate_key(version=version, encoded=encoded)

    def xpublic_key(
        self, version: Union[bytes, int] = Bitcoin.NETWORKS.MAINNET.XPUBLIC_KEY_VERSIONS.P2WPKH_IN_P2SH, encoded: bool = True
    ) -> Optional[str]:
        """
        Generates the extended public key (xpublic key) for the BIP44HD instance.

        :param version: The version bytes or integer for the xpublic key.
        :type version: Union[bytes, int]
        :param encoded: Whether to encode the xpublic key or return it in raw form, defaults to True.
        :type encoded: bool

        :return: The extended public key (xpublic key) if available, otherwise None.
        :rtype: Optional[str]
        """

        return super(BIP44HD, self).xpublic_key(version=version, encoded=encoded)

    def address(
        self, script_address_prefix: int = Bitcoin.NETWORKS.MAINNET.SCRIPT_ADDRESS_PREFIX, **kwargs
    ) -> str:
        """
        Generates the P2WPKH-in-P2SH address for the BIP49HD instance.

        :param script_address_prefix: The script address prefix for encoding the address, defaults to the main network prefix.
        :type script_address_prefix: int

        :return: The P2WPKH-in-P2SH encoded address.
        :rtype: str
        """

        return P2WPKHInP2SHAddress.encode(
            public_key=self._public_key,
            script_address_prefix=script_address_prefix,
            public_key_type=self._public_key_type
        )
